@@ -7,6 +7,8 @@
 --
 
 local array = require "util.array";
+local jid = require "util.jid";
+local url = require "socket.url";
 
 -- Source: http://xmpp.org/registrar/formtypes.html#http:--jabber.org-network-serverinfo
 local form_layout = require "util.dataforms".new({
@@ -16,6 +18,7 @@ local form_layout = require "util.dataforms".new({
 	{ name = "feedback", var = "feedback-addresses", type = "list-multi" },
 	{ name = "sales", var = "sales-addresses", type = "list-multi" },
 	{ name = "security", var = "security-addresses", type = "list-multi" },
+	{ name = "status", var = "status-addresses", type = "list-multi" },
 	{ name = "support", var = "support-addresses", type = "list-multi" },
 });
 
@@ -23,7 +26,7 @@ local form_layout = require "util.dataforms".new({
 local admins = module:get_option_inherited_set("admins", {});
 
 local contact_config = module:get_option("contact_info", {
-	admin = array.collect( admins / function(admin) return "xmpp:" .. admin; end);
+	admin = array.collect(admins / jid.prep / function(admin) return url.build({scheme = "xmpp"; path = admin}); end);
 });
 
 module:add_extension(form_layout:form(contact_config, "result"));
